/**
 * HostIDClient
 * Host IDentification Client for JRE 1.1
 *
 * Gets the identifier of a HostIDServer running on a specified
 * network host. It also can scan all hosts on a particular
 * network for HostIDServers.
 *
 * @author Vladimir Dzhuvinov <vd@valan.net>
 * @version 1.0, 09 December 2000
 * 
 *  
 *    Copyright (C) 2000, Vladimir Dzhuvinov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


import java.net.*;
import java.io.*;
import java.lang.*;
import java.util.*;


public class HostIDClient {

	// default values
	static String id = "anonymous"; 	// client id
	static String address = "localhost"; 	// remote server address
	static int port = 6800; 		// remote server port
	static int socketReadTimeout = 5000; 	// 5000 milliseconds or 5 seconds
	static int maxThreads = 32; 		// max number of client threads
	static boolean verbose = false; 	// verbose output option
	static boolean help = false; 		// help option

	// array of IP addresses to scan for HostID servers
	static InetAddress[] IPAddress;
	
	// thread counter
	static int threadCount;
	
	
	
	public static void parseArgs(String[] args) {

		// go through each argument	
		for (int i = 0; i < args.length; i++) {
		
			// parse for client id
			if (args[i].startsWith("-I")) {
				if (args[i].length() > 2)
					id = args[i].substring(2);
				else
					id = args[++i];
			}
			
			// parse for server port
			else if (args[i].startsWith("-p")) {
				try {
					if (args[i].length() > 2)
						port = Integer.parseInt(args[i].substring(2));
					else
						port = Integer.parseInt(args[++i]);
				}
				catch (NumberFormatException e) {
					System.err.println("HostIDClient: Error: Port not an integer");
					System.exit(1);
				}
			}
			
			// parse for timeout value
			else if (args[i].startsWith("-t")) {
				try {
					if (args[i].length() > 2)
						socketReadTimeout = Integer.parseInt(args[i].substring(2));
					else
						socketReadTimeout = Integer.parseInt(args[++i]);
				}
				catch (NumberFormatException e) {
					System.err.println("HostIDClient: Error: Timeout not an integer");
					System.exit(1);
				}
			}
			
			// parse for number of maximum threads
			else if (args[i].startsWith("-m")) {
				try {
					if (args[i].length() > 2)
						maxThreads = Integer.parseInt(args[i].substring(2));
					else
						maxThreads =  Integer.parseInt(args[++i]);
				}
				catch (NumberFormatException e) {
					System.err.println
					("HostIDClient: Error: Max threads not an integer");
					System.exit(1);
				}
			}
			
			// parse for verbose option switch
			else if (args[i].equals("-v") || args[i].equals("--verbose"))
				verbose = true;
			
			// parse for help option switch
			else if (args[i].equals("-h") || args[i].equals("--help"))
				help = true;
			
			// address doesn't need a switch
			else address = args[i];
			
		}
	}
	


	public static void printHelpScreen() {
	
		System.out.print
		("HostIDClient version 1.0 for Java Runtime Environment 1.2\n" +
		"\n" +
		"Gets the identifier of a HostIDServer running on a specified\n" +
		"network host. It also can scan all hosts on a particular\n" +
		"network for HostIDServers.\n" +
		"\n" +
		"Synopsis:\n" +
		"java HostIDClient [-I ID] [ADDRESS[/SUBNET-MASK]] [-p PORT]\n" +
		"                  [-t TIMEOUT] [-m MAXTHREADS] [-v, --verbose]\n" + 
		"                  [-h, --help]\n" +
		"\n" +
		"Command line options:\n" +
		"-I ID                  A string that identifies the client\n" +
		"                       when quering a HostIDServer. Overrides\n" +
		"                       default value \"anonymous\".\n" +
		"ADDRESS[/SUBNET-MASK]  Hostname or IP address to query for a\n" +
		"                       HostIDServer. A subnet mask(network\n" +
		"                       prefix-length) can be specified to scan\n" +
		"                       all hosts on that network for HostID\n" +
		"                       Servers. The default value is \"localhost\"\n" +
		"-p PORT                The port where the HostIDServer is\n" +
		"                       listening to. Overrides default port\n" +
		"                       6800.\n" +
		"-t TIMEOUT             The socket read timeout in milliseconds.\n" +
		"                       This is the amount of time the client\n" +
		"                       will wait for the HostIDServer to\n" +
		"                       respond before closing the connection.\n" +
		"                       Overrides the default value of 5000ms.\n" +
		"-m MAXTHREADS          Maximum allowed number of client threads.\n" +
		"                       Overrides default maximum of 32.\n" +
		"-v, --verbose          Produces verbose output.\n" +
		"-h, --help             Prints this help screen.\n");
		
	}
	
	
	
	public static void composeIPAddressArray () {
	
		// default bitmask covers all IP address bits
		int bitmaskLength = 32; 
		
		// check whether a subnet mask was specified
		int slashPosition = address.indexOf('/');
		
		if (slashPosition != -1) {
		
			// parse for bitmask length
			try {
				bitmaskLength = Integer.parseInt(address.substring(slashPosition+1));
			}
			catch (NumberFormatException e) {
				System.err.println("HostIDClient: Error: " +
				"prefix (subnet mask) length not an integer");
				System.exit(1);
			}
			
			// trim slash and subnet bitmask from address
			address = address.substring(0, slashPosition);
		}
		
		
		// get the four corresponding IP address bytes,
		// InetAddress.getByName() does a DNS lookup if necessary
		
		byte[] IPAddressBytes = null; 
		
		try {
			IPAddressBytes = InetAddress.getByName(address).getAddress();
		}
		catch (UnknownHostException e) {
			System.err.println("HostIDClient: Error: unknown host: " + address);
			System.exit(1);
		}
		
		// convert byte array into int array
		// and offset bytes according to their position in the IP address
		
		int[] byteShifter = new int[4];
		
		for (int i=0; i < 4; i++) {
		
			byteShifter[i] = IPAddressBytes[i];
			
			// do some ugly shifting to compensate for 
			// Java's missing unsigned
			byteShifter[i] = byteShifter[i] << 24; // drop leading '1's if any
			byteShifter[i] = byteShifter[i] >>> (i*8); // adjust byte position
			
		}
		
		// OR offset bytes to combine them 
		// into a single binary IP address
		int binaryIPAddress = 
		byteShifter[0] | byteShifter[1] | byteShifter[2] | byteShifter[3];
		
		// create subnet bitmask by looking at its length
		int bitmask = -1 << (32 - bitmaskLength);
				
		// apply bitmask to clear all host bits from the IP address
		binaryIPAddress = binaryIPAddress & bitmask;
		
		// set the IP address array size
		IPAddress = new InetAddress[1 << (32 - bitmaskLength)];
		
		// assemble the IP address array
		
		int tmp; // used for ugly shifting
		String buf; // buffer for String manipulation
		
		for (int i = 0; i < IPAddress.length; i++) {
		
			tmp = (binaryIPAddress | i); // set network and host bits
			
			// assemble IP address String in dotted form,
			// more ugly shifting
			buf = Integer.toString(tmp >>> 24) + 
			"." +
			Integer.toString((tmp << 8) >>> 24) + 
			"." +
			Integer.toString((tmp << 16) >>> 24) + 
			"." +
			Integer.toString((tmp << 24) >>> 24);
			
			try {
				IPAddress[i] = InetAddress.getByName(buf);
			}
			catch (UnknownHostException e) {
			
				// this really shouldn't happen here...
				System.err.println("HostIDClient: IP address array error");
				System.exit(1);
			
			}
		}
	}
	
	
	
	// static inner class
	public static class ClientThread extends Thread {
		
		// reference to server inet address
		InetAddress addr;
				
		public ClientThread (InetAddress addr) {
			
			// initialise reference 
			this.addr = addr;
			
			// increment thread counter
			threadCount++; 
		}
		
		public void run() {
		
			Socket socket = null;
			
			// attemp to create a new socket to specified address and port
			try {
				socket = new Socket(addr, port);

				try {	
					// send bytes ASAP
					socket.setTcpNoDelay(true);
				
					// set the socket read timeout delay, in milliseconds
					socket.setSoTimeout(socketReadTimeout);
				}
				catch (SocketException e) {
					System.err.println("HostIDServer: Error " + e.getMessage());
				}
				
				try {
					// create buffer to read input
					BufferedReader in = 
					new BufferedReader
					(new InputStreamReader(socket.getInputStream()));
	
					// create output writer that 
					// automatically flushes the output stream buffer
					PrintWriter out = 
					new PrintWriter(new BufferedWriter
					(new OutputStreamWriter(socket.getOutputStream())),true);
    
	    				// start communication with HostID server
    					String str = in.readLine();
					
					// check whether we connected to a HostID server
					if (str.equals("HostIDServer")){
				
						out.println("HostIDClient"); // say hello to HostID server
						
						str = in.readLine(); // read server ID
						
						out.println(id); // send own ID
						
						// print message that a HostID server was found
						System.out.println
						("HostIDClient: Connected to HostIDServer[" + 
						str + "] at " + socket.getInetAddress());
					}
				}
				catch (NullPointerException e) {
					if (verbose == true)
						System.out.println
						("HostIDClient: Null response from server at" +
						addr +
						", possibly not a HostID Server");
				}
				catch (InterruptedIOException e) {
					if(verbose == true)
						System.out.println("HostIDClient: " +
						"Socket read operation from " +
						addr +
						" timed out after " +
						socketReadTimeout + 
						" ms");
				}
				catch (IOException e) {
					if (verbose == true) 
						System.out.println("HostIDClient: " + e.getMessage());
				}
				finally {
					try {
						socket.close(); // close socket
					}
					catch (IOException e) {
						System.err.println("HostIDClient: " + e.getMessage());
					}
				}
			}
			catch (UnknownHostException e) {
				if (verbose == true)
					System.out.println("HostIDClient: " + 
					addr + " : " + e.getMessage() );
			}
			catch (IOException e) {
				if (verbose == true)
					System.out.println("HostIDClient: " + 
					addr + 
					" : " + 
					e.getMessage());
			}
			
			threadCount--; // decrement thread counter
		}
	}
	


	// M A I N   P R O G R A M
		

	public static void main(String[] args) {
        	
		// parse command line arguments
		parseArgs(args);
		
		// print help screen and exit if help option was selected
		if (help == true) {
			printHelpScreen();
			System.exit(0);
		}
		
	
		//compose array of IP addresses to scan
		composeIPAddressArray ();
		
		// launch client threads
		for (int i = 0; i < IPAddress.length; i++) {
		
			// check number of running threads
			while (threadCount > maxThreads) {
				try {
					Thread.currentThread().sleep(100); // sleep 100ms
				}
				catch (InterruptedException e) { /* ignore */ }
			}
			
			// create a new thread
			ClientThread client = new ClientThread(IPAddress[i]);
			
			// start thread
			client.start();

		}
	}
}
