/**
 * HostIDServer
 * Host IDentification Server for JRE 1.1
 *
 * Provides identification to a network host that is not tied to
 * a particular IP address or host name.
 *
 * @author Vladimir Dzhuvinov <vd@valan.net>
 * @version 1.0, 09 December 2000
 * 
 *  
 *    Copyright (C) 2000, Vladimir Dzhuvinov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


import java.io.*;
import java.net.*;
import java.lang.*;
import java.util.*;
import java.text.*;



public class HostIDServer {  
	
	
	// default values
	static String id = "anonymous"; 	// server id
	static int port = 6800; 		// port to listen to
	static int socketReadTimeout = 5000; 	// 5000 milliseconds or 5 seconds
	static boolean verbose = false; 	// verbose output option
	static boolean help = false; 		// help option

	
	public static void parseArgs(String[] args) {

		// go through each command line argument	
		for (int i = 0; i < args.length; i++) {
		
			// parse for client id
			if (args[i].startsWith("-I")) {
				if (args[i].length() > 2)
					id = args[i].substring(2);
				else
					id = args[++i];
			}
			
			// parse for server port
			else if (args[i].startsWith("-p")) {
				try {
					if (args[i].length() > 2)
						port = Integer.parseInt(args[i].substring(2));
					else
						port = Integer.parseInt(args[++i]);
				}
				catch (NumberFormatException e) {
					System.err.println("HostIDServer: Error: Port not an integer");
					System.exit(1);
				}
			}
			
			// parse for timeout value
			else if (args[i].startsWith("-t")) {
				try {
					if (args[i].length() > 2)
						socketReadTimeout = Integer.parseInt(args[i].substring(2));
					else
						socketReadTimeout = Integer.parseInt(args[++i]);
				}
				catch (NumberFormatException e) {
					System.err.println
					("HostIDServer: Error: Timeout value not an integer");
					
					System.exit(1);
				}
			}
			

			
			// parse for verbose option switch
			else if (args[i].equals("-v") || args[i].equals("--verbose"))
				verbose = true;
			
			// parse for help option switch
			else if (args[i].equals("-h") || args[i].equals("--help"))
				help = true;
			
			
		}
	}
	


	public static void printHelpScreen() {
	
		System.out.print
		("HostIDServer version 1.0 for Java Runtime Environment 1.2\n" +
		"\n" +
		"Provides identification to a network host that is not tied to\n" +
		"a particular IP address or host name.\n" +
		"\n" +
		"Synopsis:\n" +
		"java HostIDServer [-I ID] [-p PORT] [-t TIMEOUT]\n" +
		"                  [-v, --verbose ] [-h --help]\n" +
		"\n" +
		"Command line options:\n" +
		"-I ID           A string that identifies the HostIDServer during\n" + 
		"                a query from a HostIDClient. Overrides default\n" +
		"                value \"anonymous\".\n" +
		"-p PORT         The port where the HostIDServer is listening to.\n" +
		"                Overrides default port 6800.\n" +
		"-t TIMEOUT      The socket read timeout in milliseconds. This\n" +
		"                is the amount of time the server will wait for\n" + 
		"                the HostIDClient to respond before closing the\n" + 
		"                connection. Overrides the default value of 5000ms.\n" +
		"-v, --verbose   Produces verbose output.\n" +
		"-h, --help      Prints this help screen.\n");
		
	}
	
	
	public static void main(String[] args) {
	
		// parse command line arguments
		parseArgs(args);
		
		// print help screen
		if (help == true) {
			printHelpScreen();
			System.exit(0);
		}
	
		// for logging server messages
		DateFormat timestamp;
		
		// server socket to listen for connection
		ServerSocket s = null;
		
		// socket for stream communication
		Socket socket = null;
		
		try {
			s = new ServerSocket(port);
		}
		catch (IOException e) {
			System.err.println("HostIDServer: Error creating server socket: " + 
			e.getMessage());
			
			System.exit(1);
		}
		
		System.out.println("HostIDServer: Started " + s);

		while (true) {

			// block until a connection occurs
			try {
				socket = s.accept();
				
				try {	
					// send bytes ASAP
					socket.setTcpNoDelay(true);
				
					// set the socket read timeout delay, in milliseconds
					socket.setSoTimeout(socketReadTimeout);
				}
				catch (SocketException e) {
					System.err.println("HostIDServer: Error " + e.getMessage());
				}
				
				// start communication once connection is established
				try {
					// input stream from client     
					BufferedReader in = new BufferedReader
					(new InputStreamReader(socket.getInputStream()));
					
					// output stream to client, buffer auto flushed
					PrintWriter out = new PrintWriter
					(new BufferedWriter
					(new OutputStreamWriter(socket.getOutputStream())),true);
             
					out.println("HostIDServer"); // say hello
					
					String str = in.readLine(); // read client response
					
					// if we've got a HostIDClient exchange IDs
					if (str.equals("HostIDClient")) {
						out.println(id); // send own ID
						str = in.readLine(); // get client ID
						
						timestamp = 
						DateFormat.getDateTimeInstance
						(DateFormat.SHORT, DateFormat.SHORT);	
							
						System.out.println("HostIDServer: " +
						timestamp.format(new Date()) + 
						": HostIDClient[" + 
						str + 
						"] connection from " + 
						socket.getInetAddress()); 
						
					}
				}
				catch (NullPointerException e) {
					if (verbose == true)
						System.err.println
						("HostIDServer: Error reading data from client, " +
						"possibly not a HostID Client");
				}
				catch (InterruptedIOException e) {
					if(verbose == true)
						System.err.println("HostIDServer: " +
						socket.getInetAddress() +
						" : Socket read operation timed out after " +
						socketReadTimeout + " milliseconds");
				}
				finally {
					try {
						socket.close(); // close socket when done
					}
					catch (IOException e) {
						System.err.println("HostIDServer: " + e.getMessage());
					}
				}
			}
			catch (IOException e ) {
				if (verbose == true)
					System.err.println("HostIDServer: " + e.getMessage());
			}
		}		
	}	
}

